/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * Authors: Travis Reitter <treitter@gmail.com>
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <hildon/hildon.h>
#include <rtm-glib/rtm-glib.h>

#include "milk-main-window.h"
#include "milk-auth.h"
#include "milk-task-model.h"

G_DEFINE_TYPE (MilkMainWindow, milk_main_window, HILDON_TYPE_WINDOW)

/* less expensive than G_TYPE_INSTANCE_GET_PRIVATE */
#define MILK_MAIN_WINDOW_PRIVATE(o) ((MILK_MAIN_WINDOW ((o)))->priv)

#define NEW_TASK_PLACEHOLDER_TEXT "Enter a new task..."

static GtkWidget *default_window = NULL;

struct _MilkMainWindowPrivate
{
        MilkAuth *auth;

        GtkWidget *app_menu;

        GtkWidget *main_vbox;

        GtkWidget *new_task_entry;
        GtkWidget *task_view;
        GtkWidget *task_selector;
};

enum {
        TASK_VIEW_COLUMN_TITLE,
        N_VIEW_COLUMNS
};

typedef struct {
        const char *display_name;
        const char *id;
        gpointer    callback;
} MenuItem;

static void
milk_main_window_get_property (GObject    *object,
                               guint       property_id,
                               GValue     *value,
                               GParamSpec *pspec)
{
        switch (property_id)
        {
                default:
                        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id,
                                        pspec);
        }
}

static void
milk_main_window_set_property (GObject      *object,
                               guint         property_id,
                               const GValue *value,
                               GParamSpec   *pspec)
{
        switch (property_id)
        {
                default:
                        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id,
                                        pspec);
        }
}

static void
milk_main_window_dispose (GObject *object)
{
        G_OBJECT_CLASS (milk_main_window_parent_class)->dispose (object);
}

static void
new_task_clicked_cb (GtkButton      *button,
                     MilkMainWindow *window)
{
        g_debug ("FIXME: implement 'new task' action");
}

/* XXX: The latency between clicking "complete" and actually removing the task
 * from the view after polling the server is very long, so there's an obvious
 * lag -- it will be completely transparent (and look very fast) as soon as
 * we've got a cache in place */
static void
complete_clicked_cb (GtkButton      *button,
                     MilkMainWindow *window)
{
        MilkMainWindowPrivate *priv;
        GList *rows;
        GtkTreeModel *model;
        char *timeline;
        GError *error = NULL;

        priv = MILK_MAIN_WINDOW_PRIVATE (window);

        rows = hildon_touch_selector_get_selected_rows (
                        HILDON_TOUCH_SELECTOR (priv->task_view),
                        TASK_VIEW_COLUMN_TITLE);
        model = hildon_touch_selector_get_model (
                        HILDON_TOUCH_SELECTOR (priv->task_view),
                        TASK_VIEW_COLUMN_TITLE);

        timeline = milk_auth_timeline_create (priv->auth, &error);

        if (error) {
                g_warning (G_STRLOC ": failed to create a timeline: %s",
                           error->message);
                g_clear_error (&error);
        } else {
                while (rows) {
                        GtkTreeIter iter;
                        RtmTask *task;

                        gtk_tree_model_get_iter (model, &iter, rows->data);
                        gtk_tree_model_get (model, &iter,
                                        MILK_TASK_MODEL_COLUMN_TASK, &task,
                                        -1);

                        milk_auth_task_complete (priv->auth, timeline, task,
                                        &error);
                        if (error != NULL) {
                                g_warning (G_STRLOC ": failed to complete task "
                                                "%s: %s",
                                                rtm_task_get_id (task),
                                                error->message);
                                g_clear_error (&error);
                        }

                        rows = g_list_delete_link (rows, rows);
                }
        }
}

/* XXX: high latency until we have a cache; see the note for
 * complete_clicked_cb() */
static void
delete_clicked_cb (GtkButton      *button,
                   MilkMainWindow *window)
{
        MilkMainWindowPrivate *priv;
        GList *rows;
        GtkTreeModel *model;
        char *timeline;
        GError *error = NULL;

        priv = MILK_MAIN_WINDOW_PRIVATE (window);

        rows = hildon_touch_selector_get_selected_rows (
                        HILDON_TOUCH_SELECTOR (priv->task_view),
                        TASK_VIEW_COLUMN_TITLE);
        model = hildon_touch_selector_get_model (
                        HILDON_TOUCH_SELECTOR (priv->task_view),
                        TASK_VIEW_COLUMN_TITLE);

        timeline = milk_auth_timeline_create (priv->auth, &error);

        if (error) {
                g_warning (G_STRLOC ": failed to create a timeline: %s",
                           error->message);
                g_clear_error (&error);
        } else {
                while (rows) {
                        GtkTreeIter iter;
                        RtmTask *task;

                        gtk_tree_model_get_iter (model, &iter, rows->data);
                        gtk_tree_model_get (model, &iter,
                                        MILK_TASK_MODEL_COLUMN_TASK, &task,
                                        -1);

                        milk_auth_task_delete (priv->auth, timeline, task,
                                        &error);
                        if (error != NULL) {
                                g_warning (G_STRLOC ": failed to delete task "
                                                "%s: %s",
                                                rtm_task_get_id (task),
                                                error->message);
                                g_clear_error (&error);
                        }

                        rows = g_list_delete_link (rows, rows);
                }
        }
}

static void
priority_plus_clicked_cb (GtkButton      *button,
                          MilkMainWindow *window)
{
        g_debug ("FIXME: implement 'priority plus' action");
}

static void
priority_minus_clicked_cb (GtkButton      *button,
                           MilkMainWindow *window)
{
        g_debug ("FIXME: implement 'priority minus' action");
}

static void
edit_clicked_cb (GtkButton      *button,
                 MilkMainWindow *window)
{
        g_debug ("FIXME: implement 'edit' action");
}

static MenuItem menu_items_always_shown[] = {
        {"New Task",   "menu-item-new-task",       new_task_clicked_cb},
};

static MenuItem menu_items_selection_required[] = {
        {"Edit",       "menu-item-edit",           edit_clicked_cb},
        {"Priority +", "menu-item-priority_plus",  priority_plus_clicked_cb},
        {"Priority -", "menu-item-priority_minus", priority_minus_clicked_cb},
        {"Complete",   "menu-item-complete",       complete_clicked_cb},
        {"Delete",     "menu-item-delete",         delete_clicked_cb},
};

static void
new_task_entry_activated_cb (GtkEntry       *entry,
                             MilkMainWindow *window)
{
        MilkMainWindowPrivate *priv;
        char *name;

        priv = MILK_MAIN_WINDOW_PRIVATE (window);

        name = g_strdup (gtk_entry_get_text (entry));

        /* Strip the contents of leading and trailing whitespace, and add as a
         * new task if the result is non-empty */
        if (g_strcmp0 (g_strstrip (name), "")) {
                char *timeline;
                GError *error = NULL;

                timeline = milk_auth_timeline_create (priv->auth, &error);

                if (error) {
                        g_warning (G_STRLOC ": failed to create a timeline: %s",
                                error->message);
                        g_clear_error (&error);
                } else {
                        RtmTask *task;

                        task = milk_auth_task_add (priv->auth, timeline, name,
                                        &error);
                        if (task) {
                                /* empty out the entry and show its placeholder
                                 * text */
                                gtk_entry_set_text (entry, "");
                                gtk_widget_grab_focus (priv->task_view);
                        } else {
                                g_warning (G_STRLOC ": failed to add task: %s",
                                                error->message);
                                g_clear_error (&error);
                        }
                }
        }

        g_free (name);
}

static gboolean
new_task_entry_key_press_event_cb (GtkEntry       *entry,
                                   GdkEventKey    *event,
                                   MilkMainWindow *window)
{
        MilkMainWindowPrivate *priv;

        priv = MILK_MAIN_WINDOW_PRIVATE (window);

        if (!event || event->type != GDK_KEY_PRESS) {
                return FALSE;
        }

        switch (event->keyval) {
                case GDK_KP_Enter:
                case GDK_Return:
                        new_task_entry_activated_cb (entry, window);
                        return TRUE;
        }

        return FALSE;
}

static void
task_view_selection_changed_cb (HildonTouchSelector *view,
                                gint                 column,
                                MilkMainWindow      *window)
{
        MilkMainWindowPrivate *priv;
        GList *rows;
        gboolean show = FALSE;
        gint i;

        priv = MILK_MAIN_WINDOW_PRIVATE (window);

        rows = hildon_touch_selector_get_selected_rows (view, column);
        show = (g_list_length (rows) > 0);

        for (i = 0; i < G_N_ELEMENTS (menu_items_selection_required); i++) {
                GtkWidget *w;

                w = g_object_get_data (
                                G_OBJECT (priv->app_menu),
                                menu_items_selection_required[i].id);

                if (show)
                        gtk_widget_show (w);
                else
                        gtk_widget_hide (w);
        }

        g_list_free (rows);
}

static GtkWidget*
create_menu (gpointer user_data)
{
        HildonAppMenu *menu;
        MenuItem *menu_array;
        gint i, length;
        GtkWidget *w;

        menu = HILDON_APP_MENU (hildon_app_menu_new ());

        menu_array = menu_items_always_shown;
        length = G_N_ELEMENTS (menu_items_always_shown);
        for (i = 0; i < length; i++) {
                w = hildon_button_new_with_text (
                                HILDON_SIZE_FINGER_HEIGHT | HILDON_SIZE_AUTO_WIDTH,
                                HILDON_BUTTON_ARRANGEMENT_VERTICAL,
                                _(menu_array[i].display_name), "");
                g_signal_connect (w, "clicked",
                                G_CALLBACK (menu_array[i].callback), user_data);
                g_object_set_data (G_OBJECT (menu), menu_array[i].id, w);
                hildon_app_menu_append (menu, GTK_BUTTON (w));
                gtk_widget_show (w);
        }

        menu_array = menu_items_selection_required;
        length = G_N_ELEMENTS (menu_items_selection_required);
        for (i = 0; i < length; i++) {
                w = hildon_button_new_with_text (
                                HILDON_SIZE_FINGER_HEIGHT | HILDON_SIZE_AUTO_WIDTH,
                                HILDON_BUTTON_ARRANGEMENT_VERTICAL,
                                menu_array[i].display_name, "");
                g_signal_connect (w, "clicked",
                                G_CALLBACK (menu_array[i].callback), user_data);
                g_object_set_data (G_OBJECT (menu), menu_array[i].id, w);
                hildon_app_menu_append (menu, GTK_BUTTON (w));
                gtk_widget_hide (w);
        }

        gtk_widget_show (GTK_WIDGET (menu));

        return GTK_WIDGET (menu);
}

static void
contact_column_render_func (GtkCellLayout   *cell_layout,
                            GtkCellRenderer *renderer,
                            GtkTreeModel    *model,
                            GtkTreeIter     *iter,
                            gpointer         user_data)
{
        RtmTask *task;

        gtk_tree_model_get (
                        model, iter, MILK_TASK_MODEL_COLUMN_TASK, &task, -1);
        g_object_set (renderer, "text", rtm_task_get_name (task), NULL);

        g_object_unref (task);
}

static gboolean
begin_auth_idle (MilkMainWindow *window)
{
        MilkMainWindowPrivate *priv;

        priv = MILK_MAIN_WINDOW_PRIVATE (window);

        milk_auth_log_in (priv->auth);

        return FALSE;
}

static void
milk_main_window_constructed (GObject* object)
{
        MilkMainWindow *self = MILK_MAIN_WINDOW (object);
        MilkMainWindowPrivate *priv = MILK_MAIN_WINDOW_PRIVATE (object);
        GtkWidget *w;
        GtkTreeModel *model;
        GtkCellRenderer *renderer;
        HildonTouchSelectorColumn *col;

        w = gtk_vbox_new (FALSE, HILDON_MARGIN_DEFAULT);
        gtk_container_add (GTK_CONTAINER (self), w);
        priv->main_vbox = w;

        /*
         * New Task entry
         */
        w = hildon_entry_new (HILDON_SIZE_FINGER_HEIGHT);
        gtk_box_pack_start (GTK_BOX (priv->main_vbox), w, FALSE, FALSE, 0);

        /* FIXME: change this to hildon_gtk_entry_set_placeholder_text() is
         * fixed, since this is deprecated */
        hildon_entry_set_placeholder (HILDON_ENTRY (w),
                        _(NEW_TASK_PLACEHOLDER_TEXT));
        priv->new_task_entry = w;
        g_signal_connect (G_OBJECT (w), "activate",
                        G_CALLBACK (new_task_entry_activated_cb), self);
        g_signal_connect (G_OBJECT (w), "key-press-event",
                        G_CALLBACK (new_task_entry_key_press_event_cb), self);

        /*
         * Task List
         */
        priv->auth = milk_auth_get_default ();
        model = GTK_TREE_MODEL (milk_task_model_new (priv->auth));
        w = hildon_touch_selector_new ();


        renderer = gtk_cell_renderer_text_new ();
        g_object_set (renderer,
                        "ellipsize", PANGO_ELLIPSIZE_END,
                        NULL);

        col = hildon_touch_selector_append_column (
                        HILDON_TOUCH_SELECTOR (w), model, NULL, NULL);
        gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (col), renderer, TRUE);
        gtk_cell_layout_set_cell_data_func (
                        GTK_CELL_LAYOUT (col), renderer,
                        (GtkCellLayoutDataFunc) contact_column_render_func,
                        self, NULL);
        g_object_unref (model);

        hildon_touch_selector_set_column_selection_mode (
                        HILDON_TOUCH_SELECTOR (w),
                        HILDON_TOUCH_SELECTOR_SELECTION_MODE_MULTIPLE);
        hildon_touch_selector_set_hildon_ui_mode (
                        HILDON_TOUCH_SELECTOR (w), HILDON_UI_MODE_EDIT);
        hildon_touch_selector_unselect_all (
                        HILDON_TOUCH_SELECTOR (w), TASK_VIEW_COLUMN_TITLE);

        gtk_box_pack_start (GTK_BOX (priv->main_vbox), w, TRUE, TRUE, 0);
        g_object_set (w, "can-focus", TRUE, NULL);
        gtk_widget_grab_focus (w);

        g_signal_connect (
                        G_OBJECT (w), "changed",
                        G_CALLBACK (task_view_selection_changed_cb), self);
        priv->task_view = w;

        priv->app_menu = create_menu (self);
        hildon_window_set_app_menu (
                        HILDON_WINDOW (self), HILDON_APP_MENU (priv->app_menu));

        /* break a cyclical dependency by doing this after the window is
         * constructed */
        g_idle_add ((GSourceFunc) begin_auth_idle, self);
}

static void
milk_main_window_class_init (MilkMainWindowClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        g_type_class_add_private (klass, sizeof (MilkMainWindowPrivate));

        object_class->get_property = milk_main_window_get_property;
        object_class->set_property = milk_main_window_set_property;
        object_class->constructed = milk_main_window_constructed;
        object_class->dispose = milk_main_window_dispose;
}

static void
milk_main_window_init (MilkMainWindow *self)
{
        self->priv = G_TYPE_INSTANCE_GET_PRIVATE (
                        self, MILK_TYPE_MAIN_WINDOW, MilkMainWindowPrivate);
}

GtkWidget*
milk_main_window_get_default ()
{
        if (!default_window) {
                default_window = g_object_new (MILK_TYPE_MAIN_WINDOW, NULL);
        }

        return default_window;
}
